/******************************************************************/
/******************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/drawing.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $ */
/******************************************************************/
/*  Drawings are visual representations of network entities.      */
/*  Each drawing represents exactly one entity, but an entity may */
/*  be represented by more than one drawing.                      */
/*  A drawing appears on exactly one map.                         */
/*  All methods (routines) applicable to a drawing are prefixed   */
/*  by 'Drawing_'.                                                */
/*  A drawing could be implemented as an X11 widget or gadget.    */
/******************************************************************/

#include "map/Drawing.h"
#include "map/Layout.h"
#include "map/Interface.h"

/* A list of standard types, globally accessible */
StandardTypes Std_Type;

/******************************************************************/
/*  Test if two drawings are linked together, for this to be true */
/*  one must be an interface and the oher an system  */
/******************************************************************/

extern Boolean Drawing_are_linked (a, b)
	Drawing a,b;
{
	Drawing sys, interface;

	if (Drawing_isInterface (a)) {
		if (Drawing_isInterface (b))
			return False;
		interface = a;
		sys = b;
	} else if (Drawing_isInterface (b)) {
		interface = b;
		sys = a;
	} else
		return False;

	if (sys == Drawing_from_interface(interface) || sys == Drawing_to_interface(interface))
		return True;

	return False;
}

/******************************************************************/
/* Test if a drawing can be resized */
/******************************************************************/

extern Boolean Drawing_isResizable (o)
	Drawing o;
{
	if (Drawing_isClass (o))
		return False;
	if (Shape_resizable (shapeOf_entity (ENTITY(o))))
		return True;
	return False;
}

/******************************************************************/
/*  Select all of an entities drawings                            */
/******************************************************************/

extern void Entity_select_drawings (e)
	Entity e;
{
	Drawing d;
	FORALL (d, e->drawing, next)
		add_selection (d);
}

/******************************************************************/
/*  Allocate and init basic drawing struct                        */
/******************************************************************/

static Drawing basic_create ()
{
	Drawing new;

	PROC ("basic_create");

	new = MonetNew (struct drawing);
	new->select = select_none;
	new->map = NULL;
	new->label = NULL;
	new->labelState = label_default;
	new->resized = False;
	new->shown = show_normal;
	new->x = new->y = -1;  /*  Unplaced */
	new->base_width = new->base_height = 0;
	new->stretch_height = new->stretch_width = 0;
	new->extra  = extra_none;
	new->extra2 = extra2_none;
	new->topright = topright_none;
	new->label2 = NULL;
	new->shade  = shade_grey;
	new->reverse_video = False;
	new->next = NULL;
	new->entity = NULL;
	new->shape = NULL;

	new->list.x = 0;
	new->list.y = 0;
	new->list.width = 0;
	new->list.height = 0;
	new->list.label = NULL;

	new->backup.x = 0;
	new->backup.y = 0;
	new->backup.width = 0;
	new->backup.height = 0;
	new->backup.label = NULL;

	new->from = NULL;
	new->to = NULL;

	new -> submap = NULL;

	return new;
}

/*****************************************************************************/
/*  Returns the drawing of a class entity, or if it has none, creates one    */
/*  This is like Map_insert_drawing but for drawing classes                  */
/*  !!! This needs rationalised */
/*****************************************************************************/

extern Drawing Map_insert_drawing_class (map, class, x,y)
	Entity class;
	Position x,y;
	Map map;
{
	Drawing new;

	if (map == NULL) {
		sys_error ("Map is null");
		return NULL;
	}
	if (!Map_classesOnly(map)) {
		sys_error ("Map [%s] is not reserved for classes", Map_title (map));
		return NULL;
	}
	if (class == NULL)
		return NULL;

	if (class->drawing == NULL) {
		if (!(new = basic_create ()))
			return NULL;
		new -> entity = class;
		class -> drawing = new;
		Drawing_mapOf(new) = map;
		insert_drawingList(Map_elements (map), new);
		Drawing_prepare(new);
	}
	return class->drawing;
}

/***************************************************************/
/*  Show/hide empaphasize a drawing  */
/***************************************************************/

static struct { String name; ShownState state; } ShowTable[] = {
	{ "emphasis", show_emphasised },
	{ "deemphasis", show_deemphasized },
	{ "normal" , show_normal },
	{"hide" , show_hiden },
	{ NULL, show_normal },
};

/************************************************************************/
/*  Set the dimensions of an drawing  */
/*  !!! Should do bound checking here                                   */
/************************************************************************/

extern Boolean Drawing_dimension (o, x,y, w,h)
	Drawing o;
	Position x,y;
	Dimension w,h;
{
	Boolean can_resize, xc, yc, wc, hc;  /*  Changes  */

	can_resize = Drawing_isResizable (o);
	xc = (x > 0 && X(o)      != x);
	yc = (y > 0 && Y(o)      != y);
	wc = (can_resize && w > 0 && WIDTH(o)  != w);
	hc = (can_resize && h > 0 && HEIGHT(o) != h);

	if ( ! (wc || hc || xc || yc))
		return True;

	if (xc)
		X(o) = x;
	if (yc)
		Y(o) = y;
	if (wc)
		WIDTH(o)  = w;
	if (hc)
		HEIGHT(o) = h;

	if (wc || hc) {
		DRAWING(o)->resized = True;
		Drawing_prepare (o);
	}
	Map_adjust_corners (Drawing_mapOf(o), o, True);
 	Map_set_changed (Drawing_mapOf(o));
	return True;
}

/************************************************************************/
/*  Resize an drawing  */
/*  !!! Should undraw and redraw system */
/************************************************************************/

extern Boolean Drawing_resize (o, w, h)
	Drawing o;
	Dimension w, h;
{
	return Drawing_dimension (o, 0, 0, w,h);
}

/************************************************************************/
/*  Position an drawing on the map */
/*  !!! Should undraw and redraw system */
/************************************************************************/

extern Boolean Drawing_position (o, x,y)
	Drawing o;
	Position x,y;
{
	return Drawing_dimension (o, x,y, 0,0);
}

/***************************************************************/
/*   Set selection flag, and then redraw system */
/***************************************************************/

extern Boolean Drawing_isSelected (s)
	Drawing s;
{ return SELECT(s) != select_none; }

extern void Drawing_select (s, select)
	Drawing s;
	SelectState select;
{
	if (SELECT(s) != select) {
		Drawing_unmoved_undraw (s);
		SELECT(s) = select;
		Drawing_prepare (s);
		Drawing_unmoved_redraw (s);
	}
}

extern void Drawing_deSelect (s)
	Drawing s;
{
	if (SELECT(s) != select_none) {
		Drawing_unmoved_undraw (s);
		SELECT(s) = select_none;
		Drawing_prepare (s);
		Drawing_unmoved_redraw (s);
	}
}

/***************************************************************/
/*   Set emphasis flag, and then Drawing_redraw */
/***************************************************************/

extern Boolean Drawing_isEmphasized (s)
	Drawing s;
{ return SHOW(s) != show_normal; }

extern void Drawing_emphasize (s, emphasis)
	Drawing s;
	ShownState emphasis;
{
	if (SHOW(s) != emphasis) {
		Drawing_unmoved_undraw (s);
		SHOW(s) = emphasis;
		Drawing_prepare (s);
		Drawing_unmoved_redraw (s);
	}
}

extern void Drawing_deEmphasize (s)
	Drawing s;
{ 
	if (SHOW(s) != show_normal) {
		Drawing_unmoved_undraw (s);
		SHOW(s) = show_normal;
		Drawing_prepare (s);
		Drawing_unmoved_redraw (s);
	}
}

/***************************************************************/
/*  Can an drawing be moved?                                    */
/***************************************************************/

extern Boolean Drawing_isMoveable (o)
	Drawing o;
{
	return Drawing_isClass (o) == False;
}

/************************************************************/
/* Terribly important routine, decides what a visual drawing */
/* looks like depending on its current status */
/* !!! This is not implemented terribly efficiently and its */
/* !!! use is badly structured.                             */
/************************************************************/
/*  reachable   - border 'normal' else grey border. */
/*  unreachable - diagonal line across symbol */
/*  emphasised  - extra line on left */
/*  selected    - reverse video / thick border*/
/*  superclass  - extra border */
/*  first selected - extra lines on left and bottem */
/************************************************************/

/************************************************************************/
/* Prepare system for drawing */
/************************************************************************/

extern Boolean Drawing_prepare (drawing)
	Drawing drawing;
{
	Dimension       width, height;
	Map map;
	ExtraDrawing	extra = extra_default;
	Extra2Drawing	extra2 = extra2_none;
	ExtraTopRight   topright = topright_none;
	ExtraFlag	flag = flag_none;
	Shade		shade = shade_dark;
	Entity		e = ENTITY(drawing);

	map = Drawing_mapOf(drawing);
	if (map == NULL) {
		sys_error ("Map is NULL");
		return False;
	}

	if (map->drawnas == drawnas_list)
		Drawing_reset_list_label (drawing);

	/************************************************************/
	/*  Set up width and height of drawing                      */
	/************************************************************/

	/*  Set the label and base dimensions  */
	if ( ! LABEL(drawing))
		(void) Drawing_reset_label ((Drawing) drawing);

	width  = BASE_WIDTH(drawing);
	height = BASE_HEIGHT(drawing);

	if (DRAWING(drawing)->resized)
		if (WIDTH(drawing) < width + 2 || HEIGHT(drawing) < height + 2)
			DRAWING(drawing)->resized = False;

	/*  Use the status information to determine some display state  */
	/*  If no status information available then use defaults.       */
	if (Entity_isClass (e)) {
	} else if (V_STATUS(e)) {
		if (LABEL2(drawing))
			MonetFree (LABEL2(drawing));
		if (CAPACITY(e) > 0) {
			LABEL2(drawing) = MonetNewString(capacity2mega_string (CAPACITY(e)));
		} else {
			LABEL2(drawing) = NULL;
		}
		if (FUTURE(e) == future_planned) {
			topright = topright_cross;
			shade = shade_grey;
		} else {
			switch (STATUS(e)) {
			case status_critical:
				shade = shade_dark;
				flag   = flag_down;
				extra2 = extra2_crossonce;
				break;
			case status_reachable:
				shade = shade_dark;
				extra2 = extra2_highlight;
				flag   = flag_up;
				break;
			case status_invalid:
				shade = shade_grey;
				extra2 = extra2_crosstwice;
				break;
			case status_unknown:
				extra2 = extra2_none;
				if (isPingable (e))
					flag   = flag_halfmast;
				break;
			case status_unreachable:
			default:
				shade = shade_dark;
				extra2 = extra2_lowlight;
				flag   = flag_down;
				break;
			}
			switch (FUTURE(e)) {
			default:
				break;
			case future_temporary:
				topright = topright_T;
				break;
			case future_planned:
				/* NOTREACHED */
				topright = topright_P;
				break;
			case future_unknown:
				shade = shade_grey;
				break;
			case future_certain:
				break;
			}
		}
	}

	/*  Other non status stuff  */
	switch (SHOW(drawing)) {
	default:
		sys_error("Invalid emphasis_display");
		break;
	case show_emphasised:
		extra = extra_left;
		break;
	case show_hiden:   /* !!! Bad spelling */
	case show_deemphasized:
	case show_normal:
		break;
	}

	REVERSEVIDEO(drawing) = False;
	switch (SELECT(drawing)) {
	default:
		sys_error("Invalid select state");
		break;
	case select_none:
		break;
	case select_super:
		extra = extra_double_border;
		break;
	case select_first:
		shade = shade_heavy;
		extra = extra_leftbelow;
		break;
	case select_normal:
		shade = shade_heavy;
		/* !!! REVERSEVIDEO(drawing) = True; */
		extra = extra_border;
		break;
	}

	EXTRA1(drawing)  = extra;
	EXTRA2(drawing) = extra2;
	EXTRAFLAG(drawing) = flag;
	TOPRIGHT(drawing) = topright;
	SHADE(drawing)  = shade;

	/************************************************************/
	/*  Set up connecting lines  */
	/************************************************************/

	if (isInterface(e)) {
		drawing->from = Drawing_from_interface (drawing);
		drawing->to   = Drawing_to_interface (drawing);
	}

	if (DRAWING(drawing)->resized) {
		Shape_adjust (shapeOf_entity(ENTITY(drawing)), &WIDTH(drawing), &HEIGHT(drawing));
	} else {
		Shape_prepare (shapeOf_entity(ENTITY(drawing)),
			Drawing_isClass (drawing),  /*  Force a label if class  */
			map->connectors,
			width, height,
			&WIDTH(drawing), &HEIGHT(drawing),
			LABEL(drawing));
	}

	return True;
}

/************************************************************************/
/*  Find interfaces                                                     */
/*  Will return a drawing on same map or if hat fails on the parent map */
/************************************************************************/

extern Drawing Drawing_from_interface (d)
	Drawing d;
{
	Entity from;

	if (from = from_interface (ENTITY(d))) {
		Map map;
		Drawing result;
		map = Drawing_mapOf(d);
		FORALL (result, from->drawing, next)
			if (Drawing_mapOf(result) == map)
				return result;
		if (map->parent) {
			map = Drawing_mapOf(map->parent);
			FORALL (result, from->drawing, next)
				if (Drawing_mapOf(result) == map)
					return result;
		}
	}
	return NULL;
}

extern Drawing Drawing_to_interface (d)
	Drawing d;
{
	Entity to;

	if (to = to_interface (ENTITY(d))) {
		Map map;
		Drawing result;
		map = Drawing_mapOf(d);
		FORALL (result, to->drawing, next)
			if (Drawing_mapOf(result) == map)
				return result;
		if (map->parent) {
			map = Drawing_mapOf(map->parent);
			FORALL (result, to->drawing, next)
				if (Drawing_mapOf(result) == map)
					return result;
		}
	}
	return NULL;
}

/***************************************************************/
/*  Find a drawing given its entity and map                    */
/***************************************************************/

extern Drawing get_drawing (entity, map)
	Entity entity;
	Map map;
{
	Drawing d;
	FORALL (d, entity->drawing, next)
		if (Drawing_mapOf(d) == map)
			return d;
	return NULL;
}	

/***************************************************************/
/*  Find a drawing given its name, class and map               */
/***************************************************************/

extern Drawing Drawing_find (title, class, map)
	String title;
	Map map;
	Entity class;
{
	Entity entity;

	entity = string2instance (class, title);

	if (entity == NULL)
		return NULL;
	return get_drawing (entity, map);
}

/***************************************************************/
/*  Create basic drawing */
/***************************************************************/

extern Drawing Drawing_create_from_entity (entity, map, x, y, w, h)
	Entity entity;
	Map map;
	Position x, y;
	Dimension w,h;
{
	Drawing new;
	Entity class = Entity_classOf(entity);

	if (map == NULL) {
		sys_error ("Map is Null, trying to create instance");
		return NULL;
	} else if (Map_can_insert_drawing (map, class) == False) {
		char * class_title = MonetNewString (Entity_title (class));
		user_error ("Cannot insert drawings of class [%s] in map [%s]", class_title, Drawing_title (map));
		MonetFree (class_title);
		return NULL;
	}
	if (new = basic_create ()) {
		new -> entity = entity;
		new -> next = entity -> drawing;
		entity -> drawing = new;
		switch (Entity_typeOf(class)) {
		default:
			sys_error ("Unknown system type [%d]", Entity_typeOf(class));
			return NULL;
		case o_none:
		case o_interface:
		case o_system:
			break;
		case o_map:
			debug_message ("Creating drawing with map [%s]", Drawing_title (new));
			if (new -> submap = Map_get_map_for_entity (entity)) {
			} else {
				new -> submap = Map_create ();
				new -> submap -> entity = entity;
				new -> submap -> parent = new;
			}
			break;
		}
		Map_insert_drawing (map, new, x, y, (Dimension) w, (Dimension) h);
		return new;
	}
	return NULL;
}

/***************************************************************/
/*  Create an instance */
/***************************************************************/

extern Drawing Drawing_create (title, class, map, x, y, w, h)
	String title;
	Entity class;
	Map map;
	Position x, y;
	Dimension w,h;
{
	Entity entity;

	PROC ("create_instance");

	if (Entity_isClass (class) == False) {
		sys_error ("class is an instance");
		return NULL;
	}
	if (Entity_typeOf(class) == o_none) {
		user_error ("Sorry, instances of class [%s] are not allowed", Entity_title (class));
		return NULL;
	}
	if (entity = string2instance (class, title)) {
		if (get_drawing (entity, map)) {
			user_error ("Drawing [%s:%s] already exists on map [%s]",
				Entity_title (class), title, Map_title (map));
			return NULL;
		}
	} else if (!(entity = Entity_create_instance (class, title)))
		return NULL;

	
	return Drawing_create_from_entity (entity, map, x,y, w,h);
}


/***************************************************************/
/*  Create an instance by full name */
/***************************************************************/

extern Drawing Drawing_create_by_name (name, map, x, y, w,h)
	char * name;
	Map map;
	Position x;
	Position y;
	Dimension w;
	Dimension h;
{
	Entity class;
	String title;
	Boolean isclass;
	Entity entity;

	if (entity = parse_title (name, &title, &class, &isclass))
		return Drawing_create_from_entity (entity, map, x,y, w,h);

	if (class == NULL || title == NULL) {
		user_error("Can't parse title [%s]", name);
		return NULL;
	} else if (isclass) {
		user_error("Can't create drawing of class [%s]", name);
		return NULL;
	}
	return  Drawing_create (title, class, map, x, y,w,h);
}

/******************************************************************/
/*  Returns true if any part of drawing 'o' is in box [x,y][x+w,y+h] */
/******************************************************************/

#define overlap(x1,w1,x2,w2) ((x1+w1 > x2) && (x1 < x2+w2))
extern Boolean Drawing_in_box (o,x,y,w,h)
	Drawing o;
	Position x,y;
	Dimension w,h;
{
	Dimension pw, ph;
	Position  px, py;
	Map map = Drawing_mapOf(o);

	switch (map->drawnas) {
	case drawnas_list:
		px = LISTBOX(o).x;
		py = LISTBOX(o).y;
		pw = LISTBOX(o).width;
		ph = LISTBOX(o).height;
		break;
	case drawnas_map:
		px = X(o);
		py = Y(o);
		pw = WIDTH(o);
		ph = HEIGHT(o);
		break;
	default:
		sys_error ("Invalid drawnas[%d], [%s]", map->drawnas,
				Drawing_title(o));
		return False;
	}
	/*  True if there is an overlap in the x and the y  */
	return (overlap (x,w, px, pw) && overlap (y,h, py,ph));
}


/*  Returns true if [x,y] correctly mark objetc [o] */
extern Boolean Drawing_is_marked (o,px,py)
	Drawing o;
	Position px,py;
{
	Map map = Drawing_mapOf(o);

	switch (map->drawnas) {
	case drawnas_list:
		return (!outofBounds (px,py, LISTBOX(o).x, LISTBOX(o).y,
					LISTBOX(o).width, LISTBOX(o).height));
	case drawnas_map:
		return (!outofBounds (px,py, X(o), Y(o), WIDTH(o), HEIGHT(o)));
	default:
		sys_error ("Invalid drawnas[%d], [%s]", map->drawnas,
				Drawing_title(o));
		return False;
	}
	/* NOTREACHED */
}

/*****************************************************************************/
/* Test to see if an drawing is a map */
/*****************************************************************************/

extern Boolean Drawing_isMap(s)
	Drawing          s;
{
	return Entity_typeOf (ENTITY(s)) == o_map;
/*	return (s->map_stuff != (Map) 0); */
}

/****************************************************************/

/****************************************************************/

extern Boolean Drawing_isAnnotation (o)
	Drawing o;
{
	return Entity_subclass (Std_Type.annotation, ENTITY(o));
}

/***************************************************************/

/***************************************************************/

extern Boolean Drawing_isButton (o)
	Drawing o;
{
	return (Entity_classOf(ENTITY(o)) == Std_Type.button);
}

/***************************************************************/
/*  Free drawing                                               */
/***************************************************************/

extern void Drawing_free (d)
	Drawing d;
{
	totally_unmark (d);
	Map_extract_drawing (Drawing_mapOf(d), d);
	if (d->label2 != NULL)
		MonetFree (d->label2);
	if (d->list.label != NULL)
		MonetFree (d->list.label);
	if (ENTITY(d)) {
		/*  Delete from list of entities  */
		LIST_DEL(Drawing, d, ENTITY(d)->drawing, next);
		if (ENTITY(d)->drawing == NULL)
			Entity_free (ENTITY(d));
	}
	MonetFree (d);
}

/********************************************************************************/
/*  Initiation */
/*  Create a single, default class 'top'.  This must be inserted in the         */
/*  :Class map.  */
/********************************************************************************/

/* Use this routine after loading the file of Classes */
static void init_standardClasses ()
{
	Std_Type.ip_host      = Entity_find_or_create ("ip.host");
	Std_Type.ip_network   = Entity_find_or_create ("ip.network");
	Std_Type.ip_interface = Entity_find_or_create ("ip.interface");
	Std_Type.x25_host     = Entity_find_or_create ("x25.host");
	Std_Type.annotation   = Entity_find_or_create ("Annotation");
	Std_Type.interface    = Entity_find_or_create ("Interface");
	Std_Type.button       = Entity_find_or_create ("Button");

	if (Std_Type.button == NULL) {
		  sys_error ("Standard class [Button] not in Class file");
	}
	if (Std_Type.ip_host == NULL) {
		  sys_error ("Standard class [ip.host] not in Class file");
	}
	if (Std_Type.ip_network == NULL) {
		  sys_error ("Standard class [ip.network] not in Class file");
	}
	if (Std_Type.ip_interface == NULL) {
		sys_error ("Standard class [ip.interface] not in Class file");
	}
	if (Std_Type.x25_host == NULL) {
		sys_error ("Standard class [x25.host] not in Class file");
	}
	if (Std_Type.annotation == NULL) {
		sys_error ("Standard class [Annotation] not in Class file");
	}
	if (Std_Type.interface == NULL) {
		sys_error ("Standard class [Interface] not in Class file");
	}
}


/******************************************************************/
/*  Init and finish drawing module                                */
/******************************************************************/

extern void ClassDrawing_init ()
{
	init_standardClasses ();
}

extern void ClassDrawing_finish ()
{
	/* finish with standard classes */
}

/***************************************************************/
/***************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/drawing.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $ */
/***************************************************************/
/***************************************************************/

