/***************************************************************/
/***************************************************************/
/*  $Header: //gaia/group.local/computer/monet/lib/map/RCS/layout.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $ */
/***************************************************************/
/*  Strategy 1                                                 */
/*                                                             */
/*  Treat the nodes as mutually replusive particles, some of   */
/*  which are tied together by elastic bands, like coind shaped*/
/*  magnets on an icy surface, with the same pole facing up.   */
/*                                                             */
/*  All movement is limited to the universe (window).  The     */
/*  closer to the edge the harder it is to move towards it.    */
/*                                                             */
/*  Objects have mass in the horizontal and vertical directions*/
/*  basically this is the height and width of the node, it is  */
/*  used to stop overlapping and give large objects lots of    */
/*  space.                                                     */
/*                                                             */
/*  Linked nodes attract, force of attraction increases with   */
/*  distance of seperation.  Goes negative when very close.    */
/*                                                             */
/*  UnLinked nodes repel, force of attraction negative and     */
/*  gets more negative with proximity.                         */
/*                                                             */
/*  Gravity is 1 / r*r, but any equation of correct shape      */
/*  should work.                                               */
/*                                                             */
/***************************************************************/
/*  Strategy 2, not yet implemented                            */
/*                                                             */
/*  Lays out a network on a two dimensional map in a pleasing  */
/*  structure.  Does this by using the minimum number of       */
/*  overlapping lines.                                         */
/*                                                             */
/*  There are two parts to this problem, first reduce the      */
/*  number of overlapping lines to a minimum.  Secondly, lay   */
/*  the things out nicely.                                     */
/*                                                             */
/*  The strategy used reduces a network into a set of regions. */
/*  These regions themselves form a network, with the links    */
/*  defined as the boundaries between the regions.  Thus a     */
/*  heirarchy is produced with the children of a node/region   */
/*  being the nodes that make up the region.                   */
/***************************************************************/

#include "Drawing.h"
#include "Layout.h"
#include <math.h>
#define PI 3.1415926

/***************************************************************/
/*  Layout the map.                                            */
/***************************************************************/

extern void layout (map, list, num, width, height)
	DrawingList list;
	Map map;
	Dimension width, height;  /*  Size of the universe  */
	Cardinal num;
{
	Cardinal n_index, m_index;
	Drawing n,m;
	Position x,y;

	/* -- Reset force vectors  */
	FOR_DRAWINGLIST (n, list) {
		n->layout_x = 0;
		n->layout_y = 0;
	}

	/* -- Set up force vectors for all nodes */
	for (n_index = 1; n_index <= num; n_index ++) {
		n = getN_drawingList(list, n_index);
		for (m_index = n_index+1; m_index <= num; m_index ++) {
			Position vx, vy, distance, fx, fy;
			m = getN_drawingList(list, m_index);
			/* -- [vx, vy] is the vector between the centres of n and m */
			vx = (X(n)+WIDTH(n)/2 - X(m)+WIDTH(m)/2) / 4;
			vy = (Y(n)+HEIGHT(n)/2  - Y(m)+HEIGHT(m)/2) / 4;
			if (Drawing_are_linked (m, n)) {
				/* -- If m and n are directly linked then there will */
				/* -- be a force of attraction between them.         */
				/* -- This is taken as being directly proportional   */
				/* -- to th distance between them.  If they get too  */
				/* -- close, it is altered to replusion, to prevent  */
				/* -- overlapping.                                   */
				fx = -vx - WIDTH(n)  - WIDTH(m);
				fy = -vy - HEIGHT(n) - HEIGHT(m);
			} else {
				Position v;
				/* -- Things further away should have less affect */
				/* -- than those nearby, so we need to invert the */
				/* -- magnitude of [vx,vy]  */
				v  = (vx == 0) ? vy*vy  : (vx + vy*(vy/vx));
				fx = (v  == 0) ? width  : width  / v;
				v  = (vy == 0) ? vx*vx  : (vx*(vx/vy) + vy);
				fy = (v == 0)  ? height : height / v;
			}
			/* -- Add [vx,vy] to force vector for m and n  */
			/* -- This means, register the affect of m on n and v.v. */
			n->layout_x += fx;
			n->layout_y += fy;
			m->layout_x -= fx;
			m->layout_y -= fy;
		}
	}
	/* -- Set up new position from current position and force vector  */
	FOR_DRAWINGLIST (n, list) {
		Position x,y;
		Position w;

		/* --  Divide force by number of nodes */
		/* --  This prevents forces getting too big.  */
		/* --  This force has magnitude no greater than the size of the universe */
		n->layout_x /= num;
		n->layout_y /= num;

		x = X(n) + n->layout_x;
		if (x <= 0 || x >= width) {
			w = x / width;
			x = X(n) + n->layout_x / (w*width + 1);
		}
		y = Y(n) + n->layout_y;
		if (y <= 0 || y >= height) {
			w = y / height;
			y = Y(n) + n->layout_y / (w*height + 1);
		}
/*		user_inform ("[%s] v[%d,%d], [%d,%d] -> [%d,%d]", Drawing_title (n), 
				n->layout_x, n->layout_y, X(n), Y(n), x,y);
*/
		X(n) = x - WIDTH(n)/2;
		Y(n) = y - HEIGHT(n)/2;
	}
}

extern void layout_map (map, width, height, iterations)
	Map map;
	Dimension width, height;
	Cardinal iterations;
{
	Cardinal i, num;
	DrawingList list;

	if ( !map) {
		sys_error ("Map is null");
		return;
	}
	if ( ! Map_is_loaded (map)) {
		user_error ("Map is not loaded");
		return;
	}
	list = Map_elements (map);
	num = sizeOf_drawingList (list);
	if (num == 0) {
		user_error ("Map [%s] is empty", Map_title (map));
		return;
	}
	for (i = 1; i<= iterations ; ++i) {
		user_inform ("Doing iteration %u, %u still to do", i, iterations - i);
		layout (map, list, num, width, height);
	}
	user_inform ("Completed layout");
}

/***************************************************************/
/*  Initialise and finish                                      */
/***************************************************************/

extern void init_layout ()
{
}


extern void finish_layout ()
{
}

/***************************************************************/
/***************************************************************/
/*  $Header: //gaia/group.local/computer/monet/lib/map/RCS/layout.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $ */
/***************************************************************/
/***************************************************************/

