/*****************************************************************************/
/*****************************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/script.c,v 1.1 1991/07/22 11:37:27 a38 Exp a38 $ */
/*****************************************************************************/
/*  This module implements a set of standard operations on maps, that can    */
/*  be called from a monet script (or command line).  It is built upon the   */
/*  script library in monet/lib/general which supports loops, conditionals,  */
/*  pipes etc.                                                               */
/*****************************************************************************/

#include "map/Drawing.h"
#include "general/Help.h"
#include "general/Channel.h"

#include "map/Layout.h"
#include "map/Script.h"
#include "io/io.h"
#include <signal.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/ioctl.h>

/***************************************************************************/
/*  Definitions  */
/***************************************************************************/

Map CurrentMap = NULL, MarkedMap = NULL;
#define ISLOADED \
		if ( ! CurrentMap || ! Map_is_loaded(CurrentMap)) { \
			user_error ("Load a map before trying anything"); \
			RETURNSC; \
		}
#define ISLOADEDORCLASS \
		if ( ! Map_classesOnly(CurrentMap) && \
			! (CurrentMap && Map_is_loaded(CurrentMap))) { \
			user_error ("Load a map before trying anything"); \
			RETURNSC; \
		}

#define EXPECT_SELECTED(num) \
	if (numberOf_Selection () != num) { \
		if (num == 1)                \
			user_error ("You must make exactly one selection"); \
		else                                                         \
			user_error ("You must make exactly %d selections", num);\
		RETURNSC;  \
	}

static int num_args (arg)
	char ** arg;
{
	int i;
	if ( ! arg)
		return 0;
	for (i = 0; arg[i]; ++i);
	return i;
}

/****************************************************************************/
/*  Check syntax of a pinger instriuction */
/****************************************************************************/

static String parse_monitor_instruction (monitor)
	String monitor;
{
	String m;
	Boolean okay = False;

	if (monitor == NULL) {
		user_error ("Empty monitor instruction");
		return NULL;
	}
	for (m = monitor; *m != NULL; m++) {
		if (isgraph(*m)) { /* !!! */
			okay = True;
		} else {
			/* !!! Turn strings of invalid characters into spaces */
			*m = ' ';
		}
	}
	if (okay) {
		user_warning ("Monitor instruction not really checked");
		return monitor;
	}

	user_error ("Empty monitor instruction");
	return NULL;
}

/******************************************************************************/
/*  Select routines */
/******************************************************************************/

static Drawing exactly_one_drawing ()
{
	Drawing current;

	if (numberOf_Selection () != 1) {
		user_error ("You must select exactly one system");
		return NULL;
	}
	current = current_Selection ();
	free_selection ();
	return current;
}

static Entity exactly_one_entity ()
{
	Drawing current;

	if (numberOf_Selection () != 1) {
		user_error ("You must select exactly one system");
		return NULL;
	}
	current = current_Selection ();
	free_selection ();
	return ENTITY (current);
}


static Boolean at_least_one ()
{
	if (numberOf_Selection () == 0) {
		user_error ("You must select at least one system");
		return False;
	}
	return True;
}

static Entity get_type ()
{
	Drawing type;

	if ((type = current_class ()) == NULL) {
		user_error ("You must select a Class");
		return NULL;
	}
	return ENTITY(type);
}

SCRIPTCALLBACK(do_SetPingOnSystem)
	Entity sys;
	String ping, title;

	if (sys = exactly_one_entity ()) {
		title = Entity_title(sys);
		if (!isPingable (sys)) {
			user_error ("Cannot ping this class of Entity [%s]", title);
		} else {
			ping = parse_monitor_instruction (MonetJoinStringsUntilNull(arg, NULL));
			if (ping && access_ping (sys, attribute_set, attribute_ascii, ping))
				user_inform ("Set ping description of [%s] to [%s]", title, ping);
			else
				user_error ("Could not set ping description of [%s] to [%s]", title, ping);
               	}
	}
ENDSC

SCRIPTCALLBACK(do_SetPingOnClass)
	Entity class;
	String ping, title;

	if ((class = get_type ()) == NULL)
		RETURNSC;
	ping = parse_monitor_instruction (MonetJoinStringsUntilNull(arg, " "));
	title = Entity_title (class);
	if (ping && access_ping (class, attribute_set, attribute_ascii, ping))
		user_inform ("Set ping description of [%s] to [%s]", title, ping);
	else
		user_error ("Could not set ping description of [%s] to [%s]", title, ping);
ENDSC

SCRIPTCALLBACK(do_SetTitlePattern)
	Entity class;
	String pattern;

	if (class = get_type ()) {
		pattern = MonetJoinStringsUntilNull(arg, NULL);
		set_titleFormat (class, attribute_set, attribute_ascii, pattern);
	}
ENDSC

SCRIPTCALLBACK(do_Rename)
	Entity sys;
	String title;

	user_inform ("Rename");
	if (!arg || !(arg[0])) {
		user_error ("No new name given");
	} else if (sys = exactly_one_entity ())
		if (rename_entity (sys, title))
			user_inform ("Renamed to [%s]", Entity_title (sys));
ENDSC

SCRIPTCALLBACK(do_RenameClass)
	Entity sys;

	user_inform ("Rename Class");
	if (! arg || ! arg[0]) {
        	user_error ("Must give the new name");
	} else if ((sys = get_type ()) == NULL) {
		user_error ("Must select class to be renamed");
	} else if (rename_entity (sys, arg[0]) == False) {
		user_error ("Sorry, could not rename [%s]", Entity_title (sys));
	} else {
		user_inform ("Renamed Class [%s]", Entity_title (sys));
	}
ENDSC


SCRIPTCALLBACK(do_DeleteClass)
	Entity sys;
	Drawing d;

	if ((sys = get_type ()) == NULL) {
		user_error ("Must select class to be deleted");
		RETURNSC;
	}
	user_inform ("Freeing Class [%s]", Entity_title (sys));
	FORALL (d, sys->drawing, next)
		Drawing_free (d);
ENDSC


SCRIPTCALLBACK(do_SelectInstances)
	Entity class, s;

	if ((class = get_type ()) == NULL)
		RETURNSC;
	user_inform ("Selecting Instances of [%s]", Entity_title (class));
	FOR_ENTITYLIST (s, Entity_get_instances (class))
		Entity_select_drawings (s);
ENDSC

SCRIPTCALLBACK(do_ReportInstances)
	Entity class, s;

	if ((class = get_type ()) == NULL)
		RETURNSC;
	start_report ("Instances of [%s]", Entity_title (class));
	writeln_report ("Instances of [%s]", Entity_title (class));
	FOR_ENTITYLIST (s, Entity_get_instances (class))
		writeln_report ("  %s", Entity_title (s));
ENDSC


/*****************************************************************************/
/*  Load class map  */
/*****************************************************************************/

SCRIPTCALLBACK(do_class_map)
	if (!CurrentMap) {
		user_error ("No map selected");
	} else 	if (Map_classesOnly (CurrentMap)) {
		user_error ("Already a class map");
	} else if (!Map_is_saved (CurrentMap)) {
		user_error ("Current Map [%s] not saved", Map_title (CurrentMap));
	} else {
		CurrentMap -> class_map = True;
		Map_set_drawn_as (CurrentMap, drawnas_list);
		Map_request_redraw (CurrentMap);
		CurrentMap->saved = True;
	}
ENDSC

/*****************************************************************************/
/*  Select or report componented of an object */
/*****************************************************************************/

static void select_components (super, report)
	Boolean super, report;
{
	Entity sys, s;
	EntityList list;

	if (sys = exactly_one_entity ()) {
		if (super) {
			user_inform ("Selecting Supercomponents of [%s]", Entity_title (sys));
			list = Entity_supercomponents (sys);
		} else {
			user_inform ("Selecting Subcomponents of [%s]", Entity_title (sys));
			list = Entity_subcomponents (sys);
		}
		FOR_ENTITYLIST (s, list)
			if (report)
				writeln_report ("  %s", Entity_title (s));
			else
				Entity_select_drawings (s);
	}
}

/***************************************************************************/
/*  Select a set of drawings from a map and call a procedure to process    */
/*  them.  */
/***************************************************************************/

static void select_where (arg, command)
	String * arg;
	void (* command) ();
{
	String value, pattern, query;
	char attribute[1024], patbuf[1024];
	AttributeID id;
	Drawing d;
	int cc;

	query = MonetJoinStringsUntilNull (arg, NULL);
	cc = sscanf (query, "%[^ \t\n=]=%s", attribute, patbuf);
	if (cc == 2) {
		id = String2AttributeID (attribute);
		if (! id) {
			user_error ("No such attribute [%s]", attribute);
			return;
		}
		pattern = patbuf;
	} else {
		id = NULL;
		pattern = query;
	}
	if (! pattern || ! *pattern)
		pattern = "*";
	if (id)
		debug_message ("select arg[%s] [%s] = [%s]", arg[0], attribute, pattern);
	else
		debug_message ("select arg[%s] [%s]", arg[0], pattern);

	FOR_DRAWINGLIST (d, Map_elements (CurrentMap)) {
		if (id)
			value = get_stringAttribute (ENTITY(d), id);
		else
			value = Drawing_title (d);
		debug_message ("wildmat [%s] [%s]", value, pattern);
		if (value && wildmat (value, pattern))
			command (d);
	}
}

/*********************************************************************/
/* Super and Sub Components */
/*********************************************************************/

SCRIPTCALLBACK(do_SelectSupercomponent)
	Entity super;

	if (super = exactly_one_entity ())
		if (isSystem (super) || isInterface (super)) {
			Drawing d;
			user_inform ("Toggling Super component [%s]", Entity_title (super));
			FORALL (d, super->drawing, next)
				toggle_supercomponent (d);
		} else {
			user_error ("Invalid component [%s]", Entity_title (super));
		}
ENDSC

SCRIPTCALLBACK(do_MakeSubcomponent)
	Entity sub, super;
	char super_title [2048];  /* !!! */
	Drawing drawing_super;

	if (!at_least_one ())
		RETURNSC;
	drawing_super = current_supercomponent ();
	if (drawing_super == NULL)  {
		user_error ("No supercomponent selected");
		RETURNSC;
	}
	super = ENTITY(drawing_super);
	strcpy (super_title, Entity_title(super));

	FOR_SELECTED_ENTITY (sub)
		if (isSubcomponent (super, sub)) {
			user_error ("[%s] is already a subcomponent of [%s]",
				Entity_title(sub), super_title);
		} else {
			user_inform ("Making [%s] a subcomponent of [%s]",
				Entity_title(sub), super_title);
			make_subcomponent (super, sub);
		}
	END_SELECTED_ENTITY
	free_selection ();
	unmark_supercomponent (drawing_super);
ENDSC

SCRIPTCALLBACK(do_UnMakeSubcomponent)
	Entity super, sub;
	char super_title[2048];  /* !!! */
	Drawing drawing_super;

	if (!at_least_one ())
		RETURNSC;

	drawing_super = current_supercomponent ();
	if (drawing_super == NULL)  {
		user_error ("No supercomponent selected");
		RETURNSC;
	}
	super = ENTITY(drawing_super);

	strcpy (super_title, Entity_title(super));
	FOR_SELECTED_ENTITY (sub)
		if (isSubcomponent (super, sub)) {
			unmake_subcomponent (super, sub);
			user_inform ("[%s] is no longer a subcomponent of [%s]",
				Entity_title(sub), super_title);
		} else {
			user_error ("[%s] is not a subcomponent of [%s]",
				Entity_title(sub), super_title);
		}
	END_SELECTED_ENTITY	
	free_selection ();
	unmark_supercomponent (drawing_super);
ENDSC

SCRIPTCALLBACK(do_SelectSubcomponents)
	select_components (False, False);
ENDSC

SCRIPTCALLBACK(do_SelectSupercomponents)
	select_components (True, False);
ENDSC

SCRIPTCALLBACK(do_ReportSupercomponents)
	select_components (True, True);
ENDSC

SCRIPTCALLBACK(do_ReportSubcomponents)
	select_components (False, True);
ENDSC


/***************************************************************************/
/*  Full Information on specified Entities  */
/***************************************************************************/

static void info_do_command (o)
	Drawing o;
{
	writeln_report (Entity2Report(ENTITY(o), False));
}

SCRIPTCALLBACK(do_info)
	Drawing o;

	ISLOADED;
	if (arg)
		select_where (arg, info_do_command);
	else
		user_error ("info <pattern>");
ENDSC

/***************************************************************************/
/* list geometry of drawings in map  */
/***************************************************************************/

static void list_geometry_do_command (d)
	Drawing d;
{
	writeln_report ("%s %s %d %d %u %u", 
		Drawing_title (d), LABEL(d), X(d), Y(d), WIDTH(d), HEIGHT(d));
}

SCRIPTCALLBACK(do_list_geometry)
	Drawing o;

	ISLOADED;
	if (arg)
		select_where (arg, list_geometry_do_command);
	else
		user_error ("list-geometry <pattern>");
ENDSC

/***************************************************************************/
/* list connections geometry of drawings in current map  */
/***************************************************************************/

static void list_links_do_command (o)
	Drawing o;
{
	Entity entity = ENTITY(o);
	Drawing from, to;
	if (isInterface (entity))
		if ((from = Drawing_from_interface(o)) &&
			(to = Drawing_to_interface(o)))
				writeln_report ("%s %s",
					Drawing_title (from), Drawing_title (to));
}

SCRIPTCALLBACK(do_list_links)
	Drawing o;

	ISLOADED;
	if (arg)
		select_where (arg, list_links_do_command);
	else
		user_error ("list-links <pattern>");
ENDSC


/***************************************************************************/
/*  One line report on each specified Entities  */
/***************************************************************************/

static void report_do_command (o)
	Drawing o;
{
	writeln_report ("%s   %s",
		pretty_simple_status_report (V_STATUS(ENTITY(o))),
		Drawing_title (o));
}

SCRIPTCALLBACK(do_report)
	Drawing o;

	ISLOADED;
	if (arg) {
		writeln_report ("%s | Title", pretty_status_report_heading ());
		select_where (arg, report_do_command);
	} else
		user_error ("report <pattern>");
ENDSC

/***************************************************************************/
/*  Add to selection list all specified Entities  */
/***************************************************************************/

static void select_do_command (o)
	Drawing o;
{
	add_selection (o);
}

SCRIPTCALLBACK(do_select)
	Drawing o;

	ISLOADED;
	if (arg)
		select_where (arg, select_do_command);
	else
		user_error ("select <pattern>");
ENDSC

/***************************************************************************/
/*  Position and set width and height of a drawing  */
/***************************************************************************/

SCRIPTCALLBACK(do_current_map)
	if (Map_classesOnly(CurrentMap))
		result = "Class Map";
	else if ( ! CurrentMap || ! Map_is_loaded(CurrentMap))
		result = "Not Loaded";
	else
		result = Map_title (CurrentMap);
ENDSC

SCRIPTCALLBACK(do_get_edb_file)
	result =  Config_entity_database_file;
ENDSC

SCRIPTCALLBACK(do_get_monetd)
	static char host [1024]; /* !!! */
	(void) sprintf (host, "%s+%d", Config_pinger_address, Config_pinger_port);
	result = host;
ENDSC

SCRIPTCALLBACK(do_get_monetmapd)
	static char host [1024]; /* !!! */
	(void) sprintf (host, "%s+%d", Config_mapserver_address, Config_mapserver_port);
	result = host;
ENDSC

SCRIPTCALLBACK(do_list_maps)
	Map          m;
	Cardinal n = 0;
	start_report ("Maps");
	FORALLMAPS (m) {
		++n;
		write_report ("%u", n);
		if (Map_classesOnly (m)) {
			write_report (" Class Map [%u classes]",
				sizeOf_drawingList (Map_elements (m)));
		} else if (Map_is_loaded(m)) {
			if (m->saved)
				write_report (" Saved [%s]", Map_title (m));
			else
				write_report (" Unsaved [%s]", Map_title (m));
			write_report (" [%u drawings]",
				sizeOf_drawingList (Map_elements (m)));

		} else
				write_report (" UnLoaded");
		if (m == CurrentMap)
			write_report (" [Current]");
		writeln_report ("");
	}
	if (n == 0)
		writeln_report ("No maps");
ENDSC

SCRIPTCALLBACK(do_change_map)
	Map          m;
	int num;

	if (! arg || ! arg[0]) {
		user_error ("map <map name>");
	}
	num = atoi(arg[0]);
	if (num > 0) {
		Cardinal n = 0;
	 	FORALLMAPS (m) {
			++n;
			if (n == num) {
				CurrentMap = m;	
				if (Map_is_loaded(m))
					user_inform ("Current map changed to [%s]", Map_title (m));
				else
					user_inform ("New map is unloaded");
				RETURNSC;
			}
		}
	} else {
	 	FORALLMAPS (m) {
			if (Map_is_loaded(m) && EQS(Map_title (m), arg[0])) {
				CurrentMap = m;	
				user_inform ("Current map changed to [%s]", Map_title (m));
				RETURNSC;
			}
		}
	}
	user_error ("Could not find map [%s]", arg[0]);
ENDSC

static void discard_map ()
{
	user_inform ("Discarding Map");
	Map_disable_set_changed ();
	Map_discard (CurrentMap);
	Map_enable_set_changed ();
	Map_set_changed (CurrentMap);
	Map_request_redraw (CurrentMap);
}

SCRIPTCALLBACK(do_discard_map)
	ISLOADED;
	if (CurrentMap->saved)
		discard_map ();
	else
		confirm_box (discard_map, "Do you really want to discard this map?");

ENDSC


SCRIPTCALLBACK(do_free_map)
	if (sizeOf_drawingList(CurrentMap->elements) > 0) {
		sys_error("Can't free a map with sub components");
		RETURNSC;
	}
	Map_free (CurrentMap);
ENDSC


SCRIPTCALLBACK(do_insert_drawing)
	int x,y, w,h;
	Drawing drawing;
	Entity entity;
	String title;
	ISLOADED;

	if (!arg) {
		user_error ("insert drawing x y w h");
		RETURNSC;
	}
	title = arg[0];
	x = atoi (arg[1]);
	y = atoi (arg[2]);
	w = atoi (arg[3]);
	h = atoi (arg[4]);
	if (entity = Entity_find_or_create (title))
		if (drawing = get_drawing (entity, CurrentMap))
			Map_insert_drawing (CurrentMap, drawing,
				(Position) x, (Position) y, (Dimension) w, (Dimension) h);
		else
			Drawing_create_from_entity (entity, CurrentMap,
				(Position) x, (Position) y, (Dimension) w, (Dimension) h);
ENDSC

SCRIPTCALLBACK(do_get_selection)
	if (current_Selection ())
		result = Drawing_title (current_Selection ());
ENDSC

SCRIPTCALLBACK(do_explore)
	Drawing sys;

	if (!at_least_one ())
		RETURNSC;
	FOR_SELECTIONS (sys) {
		user_inform ("Exploring from [%s]", Drawing_title (sys));
		explore_connectedSystems (sys);
	}
	free_selection ();
ENDSC


SCRIPTCALLBACK(do_list_selected)
	Drawing sys;
	Cardinal id = 0;
	FOR_SELECTIONS (sys) {
		id ++;
		user_inform ("  %u %s", id, Drawing_title (sys));
	}
	user_inform ("%u Selections", id);
ENDSC

SCRIPTCALLBACK(do_SelectPinged)
	Drawing sys;
	FOR_SELECTIONS (sys)
		if (PINGID(ENTITY(sys)) == 0)
			remove_selection (sys);
ENDSC

SCRIPTCALLBACK(do_SelectPingable)
	Drawing sys;

	if (numberOf_Selection () == 0) {
		user_error ("You must select at least one system");
		RETURNSC;
	}
	FOR_SELECTIONS (sys)
		if (isPingable(ENTITY(sys)) == False)
			remove_selection (sys);
ENDSC

SCRIPTCALLBACK(do_RequestPing)
	Entity sys;
	Cardinal pinged = 0, total;

	total = numberOf_Selection ();
	FOR_SELECTED_ENTITY (sys)
		if (try_redirect (sys))
			pinged ++;
	END_SELECTED_ENTITY

	if (total == 1) {
		sys=ENTITY(current_Selection ());
		if (pinged == 0)
			user_error ("Cannot monitor [%s]", Entity_title (sys));
		else
			user_inform ("[%s] will be monitored", Entity_title (sys));
	} else
		user_inform ("%d out of the %d selected systems could be monitored", pinged, total);

	free_selection ();
ENDSC


SCRIPTCALLBACK(do_ClearPing)
	String name = arg[0];
	Entity sys;
	Cardinal cleared = 0, total;

	total = numberOf_Selection ();
	FOR_SELECTED_ENTITY (sys)
		if (stop_redirect (sys))
			cleared ++;
	END_SELECTED_ENTITY
	if (total == 0)
		user_error ("Nothing selected");
	else if (total == 1) {
		Drawing d = current_Selection ();
		user_inform ("[%s] has been cleared", Drawing_title (d));
	} else
		user_inform ("%d out of the %d selected systems where cleared", cleared, total);

	free_selection ();
ENDSC


SCRIPTCALLBACK(do_ListRedirects)
	list_redirects ();
ENDSC

SCRIPTCALLBACK(do_ReconnectPinger)
	int port;
	String host;
	char host_buf[1024]; /* !!! */
	RedirectRule rule;

	if (! arg[0]) {
		host = Config_pinger_address;
		port = Config_pinger_port;
	} else {
		int cc;
		host = host_buf;
		host_buf[0] = '\0';
		cc = sscanf (arg[0], "%[^\n \t+]%*[ +]%d", host, &port);
		if (cc < 2) {
			port = Config_pinger_port;
		}
		if (MonetStrlen (host) == 0)
			host = Config_pinger_address;
	}
	clear_redirects ();
	if (rule = create_redirect (host, port, "*")) {
		if (redirect_connected (rule)) {
			user_inform ("Connected to monitor at host [%s] port %d", host, port);
		} else {
			user_inform ("Couldn't connect to pinger at host [%s] port %d", host, port);
		}
	} else {
		sys_error ("Couldn't create redirect");
	}
ENDSC

SCRIPTCALLBACK(do_DisconnectPinger)
	clear_redirects ();
ENDSC


SCRIPTCALLBACK(do_free_selection)
	free_selection ();
ENDSC

SCRIPTCALLBACK(do_load_db)
	String name = arg[0];
	if (name && * name)
		ClassEntity_load_db (name);
	else
		ClassEntity_load_db (Config_entity_database_file);
ENDSC

SCRIPTCALLBACK(do_save_db)
	String name = arg[0];
	if (name && * name)
		ClassEntity_save_db (name);
	else
		ClassEntity_save_db (Config_entity_database_file);
ENDSC


SCRIPTCALLBACK(do_load)
	if (arg && arg[0]) {
		if (!CurrentMap)
			CurrentMap = Map_create ();
		Map_load (CurrentMap, arg[0]);
	}
ENDSC

SCRIPTCALLBACK(do_translate)
	user_inform ("Translate");
	if (! CurrentMap) {
		user_error ("No Current Map selected");
	} else if (! MarkedMap) {
		user_error ("No Marked Map");
	} else {
		(void) Map_translate (CurrentMap, MarkedMap);
	}
ENDSC

SCRIPTCALLBACK(do_save)
	ISLOADED;
	Map_save (CurrentMap);
ENDSC


SCRIPTCALLBACK(do_saveall)
	Map          m;
	FORALLMAPS(m)
		if (m->saved == False)
			(void) Map_save (m);
ENDSC

SCRIPTCALLBACK(do_Map_set_grid)
	String grid_string = arg[0];
	Cardinal grid;

	if (sscanf (grid_string, "%u", &grid) == 1) {
		if (grid > 7) {
			user_error ("Grid value is set too large [%d]", 1<<grid);
		} else {
			Map_set_grid (CurrentMap, (Dimension) grid, (Dimension) grid);
			user_inform ("The Entity grid has been set to %dx%d", 1<<grid, 1<<grid);
		}
	} else {
		user_error ("Set grid requires a cardinal number, not [%s]", grid_string);
	}
ENDSC

SCRIPTCALLBACK(do_Map_setLabel)
	user_inform ("Set Label State");
	user_error ("Set label state not yet implemented");
ENDSC

SCRIPTCALLBACK(do_Map_SetDefaultDomain)
	String name;
	user_inform ("Set Domain");
	if (name = arg[0]) {
		user_inform ("Setting domain on map [%s] to [%s]", Map_title (CurrentMap), name);
		Map_set_domain (CurrentMap, name);
	}
ENDSC

SCRIPTCALLBACK(do_Map_mark)
	if (! CurrentMap) {
		RETURNSC;
	} else if (Map_classesOnly (CurrentMap)) {
		user_error ("Can't mark a class map");
		RETURNSC;
	} else if (MarkedMap) {
		MarkedMap -> marked = False;
		Map_request_redraw (MarkedMap);
	}
	if (CurrentMap == MarkedMap) {
		MarkedMap = NULL;
	} else {
		MarkedMap = CurrentMap;
		MarkedMap -> marked = True;
	}
	Map_request_redraw (CurrentMap);
ENDSC

/******************************************************************************/
/*  Print/dump maps  */
/******************************************************************************/

SCRIPTCALLBACK(do_dump_text)
	ISLOADED;
	(void) Map_draw_text (CurrentMap);
ENDSC

SCRIPTCALLBACK(do_lpr)
	String filename;

	ISLOADED;
	if (filename = Map_draw_postscript (CurrentMap, arg[0]))
		print_postscript (filename);
ENDSC

SCRIPTCALLBACK(do_dump_ps)
	ISLOADED;
	(void) Map_draw_postscript (CurrentMap, arg[0]);
ENDSC

/******************************************************************************/
/*  Remote map server  */
/******************************************************************************/

static void get_monetmapd_address (arg, host, port)
	char ** arg, ** host;
	int *port;
{
	int cc;
	String address;
	char host_s[512]; /* !!! */
	*host = host_s;
	if (arg && arg[0])
		cc = sscanf (arg[0], "%[^ \t+]%*[\t +]%d", host_s, port);
	else
		cc = 0;
	if (cc < 2)
		*port = Config_mapserver_port;
	if (cc < 1)
		*host = Config_mapserver_address;
}

SCRIPTCALLBACK(do_remote_dir)
	String host;
	int port;
	get_monetmapd_address (arg, &host, &port);
	Map_ListRemoteDirectory (host, port);
ENDSC

SCRIPTCALLBACK(do_get_remote_map)
	String host;
	int port;
	ISLOADED;
	get_monetmapd_address (arg, &host, &port);
	read_remote_map (host, port, arg[1]);
ENDSC

SCRIPTCALLBACK(do_put_remote_map)
	String host;
	int port;
	ISLOADED;
	get_monetmapd_address (arg, &host, &port);
	save_remote_map (host, port, CurrentMap);
ENDSC

/******************************************************************************/

/******************************************************************************/

SCRIPTCALLBACK(do_save_db_remote)
	ISLOADED;
	user_error ("Not implemented yet");
ENDSC

SCRIPTCALLBACK(do_local_dir)
	Map_ListLocalDirectory (Config_map_directory);
ENDSC

SCRIPTCALLBACK(do_shapes)
	ClassShape_list ();
ENDSC

SCRIPTCALLBACK(do_attributes)
	ClassAttribute_report ();
ENDSC

SCRIPTCALLBACK(do_stat)
	if (Map_is_loaded(CurrentMap)) {
		user_inform ("Current Map is [%s] with %u drawings",
			Map_title (CurrentMap),
			sizeOf_drawingList (Map_elements (CurrentMap)));
	} else {
		user_inform ("No map loaded just now");
	}
	user_inform ("Local maps in directory [%s]", Config_map_directory);
ENDSC

SCRIPTCALLBACK(do_straighten)
	Drawing sys;

	user_inform ("Straighten");
	if (!at_least_one ())
		RETURNSC;

	FOR_SELECTIONS (sys) {
		if (position_interfaces (sys) == False) {
			user_error ("Position interfaces failed [%s]", Drawing_title (sys));
		}
	}
	free_selection ();
ENDSC


SCRIPTCALLBACK(do_join)
	Entity sys1 = NULL, sys2 = NULL, interface = NULL;
	Drawing d1, d2, d;
	Entity class;
	Map map = NULL;;

	user_inform ("Join");
	if (! CurrentMap) {
		user_error ("No map selected");
		RETURNSC;
	}
	FOR_SELECTIONS (d) {
		Map map_drawing;
		Entity sys = d->entity;
		map_drawing = Drawing_mapOf (d);
		if (Map_root (map_drawing) != CurrentMap)
			continue;
		if (map) {
			if (map_drawing != map) {
				user_error ("All selections must be on same map");
				RETURNSC;
			}
		} else
			map = map_drawing;
		if (isInterface (sys)) {
			if (interface) {
				user_error ("Only one Interface required");
				RETURNSC;
			} else {
				interface = sys;
			}
		} else if (sys1) {
			if (sys2) {
				user_error ("No more than 2 systems required");
				RETURNSC;
			} else {
				d2 = d;
				sys2 = sys;
			}
		} else {
			d1 = d;
			sys1 = sys;
		}
	}

	free_selection ();

	if (sys1 && sys2) {
		if (Drawing_isMap (d1) || Drawing_isMap(d2)) {
			user_error ("Cannot interface to a Map");
		} else if (are_interfaced ((sys1), (sys2))) {
			user_error ("These systems are already interfaced");
		} else if (interface) {
			/* !!! Should check both operations are possible before trying */
			/* !!! either. */
			make_interface (interface, sys2);
			make_interface (interface, sys1);
		} else {
			if ( ! (class = get_type ()))
				class = Entity_choose_interface (sys1,sys2);
			if (class && isInterfaceClass (class)) {
				make_anonymousInterface (class, d1, d2);
			} else {
				user_error ("Must first select an interface class");
			}
		}
	} else if (interface) {
		make_interface (interface, sys1);
	} else {
		user_error ("Must make more than one selection");
	}
ENDSC

SCRIPTCALLBACK(do_detach)
	Entity sys1, sys2;
	Drawing d1, d2;

	user_inform ("Detach");

	d1 = current_Selection ();
	if ( ! d1) {
		user_error ("Nothing selected");
		RETURNSC;
	}
	sys1 = ENTITY(d1);
	d2 = second_selection ();
	if ( ! d2) {
		if (isInterface (sys1)) {
			if (sys2 = from_interface (sys1))
				unmake_interface (sys1, sys2);
			if (sys2 = to_interface (sys1))
				unmake_interface (sys1, sys2);
			RETURNSC;
		} else {
			user_error ("Must select an interface and a system");
			RETURNSC;
		}
	}
	sys2 = ENTITY(d2);

	if (are_interfaced (sys1, sys2) == False) {
		user_error ("These systems aren't interfaced");
	} else if (isInterface (sys1)) {
		unmake_interface (sys1, sys2);
	} else if (isInterface (sys2)) {
		unmake_interface (sys2, sys1);
	} else {
		user_inform ("Sorry, you have to delete the interface box itself");
	}
	free_selection ();
ENDSC

static void delete_many ()
{
	Drawing sys;

	sys= current_Selection ();
	while (sys != NULL) {
      		user_inform ("Freeing drawing of [%s]", Drawing_title (sys));
	      	Drawing_free (sys);
		sys = current_Selection ();
	}
}

SCRIPTCALLBACK(do_delete)
	Drawing sys;

	user_inform ("Delete");
	if (sys = exactly_one_drawing ()) {
		user_inform ("Freeing Drawing [%s]", Drawing_title (sys));
		Drawing_free (sys);
	} else if (at_least_one ()) {
		confirm_box (delete_many, "You really want to Delete all Selected Objects?");
	} else {
		user_error ("Nothing selected for deletion");
	}
ENDSC

static void delete_all_interfaces ()
{
	Drawing d;

	FOR_DRAWINGLIST (d, Map_elements (CurrentMap))
		if (Drawing_isInterface (d))
		      	Drawing_free (d);
}

SCRIPTCALLBACK(do_delete_interfaces)
	ISLOADED;
	user_inform ("Delete Interfaces");
	confirm_box (delete_all_interfaces, "You really want to Delete all Interfaces?");
ENDSC

SCRIPTCALLBACK(do_selectLocalHost)
	ISLOADED;
	select_local_host (CurrentMap);
ENDSC

SCRIPTCALLBACK(do_mapset)
	String attribute, value, value2;

	ISLOADEDORCLASS;

	if ( ! arg) {
		user_error ("No arguments given to mapset");
		RETURNSC;
	}
	attribute = arg[0];
	if (! attribute)
		RETURNSC;
	value = arg[1];
	if (! value)
		RETURNSC;
	if (EQS(value, "=")) {
		value = arg[2];
		value2 = arg[3];
		if (! value)
			RETURNSC;
	} else
		value2 = arg[2];


	if (EQS(attribute, "font")) {
		CurrentMap->scale = String2Scale (value);
		Map_request_redraw (CurrentMap);
	} else if (EQS(attribute, "grid") && value2) {
		CurrentMap->grid_x = atoi(value);
		CurrentMap->grid_y = atoi(value2);
	} else if (EQS(attribute, "showload")) {
		Map_set_show_load (CurrentMap, True);
		Map_request_redraw (CurrentMap);
	} else if (EQS(attribute, "hideload")) {
		Map_set_show_load (CurrentMap, False);
		Map_request_redraw (CurrentMap);
	} else if (EQS(attribute, "drawnas")) {
		CurrentMap->drawnas = string2drawnas (value);
		Map_request_redraw (CurrentMap);
	} else if (EQS(attribute, "domain")) {
		Map_set_domain (CurrentMap, value);
		Map_request_redraw (CurrentMap);
	} else if (EQS(attribute, "background")) {
		Map_set_background_file (CurrentMap, value);
		Map_request_redraw (CurrentMap);
	} else if (EQS(attribute, "background_colour")) {
		Map_set_background_colour (CurrentMap, value);
		Map_request_redraw (CurrentMap);
	} else if (EQS(attribute, "connectors")) {
		Drawing d;
		Cardinal count = 0;
		CurrentMap -> connectors = String2ConnectorState (value);
		FOR_DRAWINGLIST (d, Map_elements (CurrentMap)) {
			if (Shape_isConnector (shapeOf_entity (ENTITY(d)))) {
				Drawing_unmoved_undraw (d);
				Drawing_prepare (d);
				Drawing_unmoved_redraw (d);
				count ++;
			}
		}
		if (count == 0)
			user_inform ("No connectors changed");
		else
			user_inform ("%u connectors changed", count);
	}
ENDSC


SCRIPTCALLBACK(do_redraw)
	if (CurrentMap)
		Map_request_redraw (CurrentMap);
ENDSC

SCRIPTCALLBACK(do_contacts)
	Entity sys;

	if (!at_least_one ())
		RETURNSC;
	FOR_SELECTED_ENTITY (sys)
		show_contact (sys);
	END_SELECTED_ENTITY
ENDSC

static void undo_layout ()
{
	Map_restore (CurrentMap);
}

SCRIPTCALLBACK(do_layout)
	String times = arg[0];
	Cardinal iterations;
	int cc;

	ISLOADED;

	if (!times)
		iterations = 1;
	else {
		cc = sscanf (times, "%u", &iterations);
		if (cc != 1 || iterations == 0)
			iterations = 1;
	}
	Map_backup (CurrentMap);
	layout_map (CurrentMap, (Dimension) 1000, (Dimension) 1000, iterations);
	Map_request_redraw (CurrentMap);
	confirm_box (undo_layout, "Do you want to undo this layout?");
ENDSC

SCRIPTCALLBACK(do_random_layout)
	int seed;
	int cc;
	Position xl,yl,xr,yr;
	Drawing d;

	ISLOADED;

	if (arg && arg[0]) {
		seed = atoi(arg[0]);
	} else {
		seed = 27;
	}
	if (Map_corners (CurrentMap, &xl,&yl, &xr,&yr, False) == False) {
		user_warning ("Map [%s] is empty", Map_title (CurrentMap));
		RETURNSC;
	}
	Map_backup (CurrentMap);
	srand (seed);
	FOR_DRAWINGLIST (d, Map_elements (CurrentMap)) {
		Position x,y;
		if ((x = rand () % (xr-xl - WIDTH(d) - 100)+50) > 0)
			X(d) = x;
		if ((y = rand () % (yr-yl - HEIGHT(d) - 100)+50) > 0)
			Y(d) = y;
		Map_adjust_corners (CurrentMap, d, True);
	}
	Map_request_redraw (CurrentMap);
	confirm_box (undo_layout, "Do you want to undo this layout?");
ENDSC

SCRIPTCALLBACK(do_action)

ENDSC

SCRIPTCALLBACK(do_set_future)
	FutureStatus future;
	Entity sys;

	future = string2future ((String) arg[0]);

	if (at_least_one ()) {
		FOR_SELECTED_ENTITY (sys)
			set_future (V_STATUS(sys), future);
		END_SELECTED_ENTITY
		free_selection ();
	}
ENDSC

SCRIPTCALLBACK(do_set_capacity)
	Entity sys;
	Capacity c;
	int cc;
	String value;

	if (!at_least_one ())
		RETURNSC;
	if (value = arg[0])
		cc = sscanf (value, "%lu", &c);
	if (!value || cc == 0) {
		user_error ("Enter capacity as an unsigned integer");
		RETURNSC;
	}
	FOR_SELECTED_ENTITY (sys)
		if (!set_capacity (V_STATUS(sys), c))
			user_error ("Sorry, could not set capacity of [%U] on [%s]", c, Entity_title (sys));
	END_SELECTED_ENTITY
	free_selection ();
ENDSC


SCRIPTCALLBACK(do_gather_info)
	Entity sys;

	if (sys = exactly_one_entity ())
		gather_generalInfo (sys);
ENDSC

SCRIPTCALLBACK(do_ReportClass)
	Entity type;
	
	if ((type = get_type ()) == NULL)
		RETURNSC;

	Entity_status_report (type);
ENDSC

SCRIPTCALLBACK(do_StatusReport)
	Entity sys;

	if (numberOf_Selection () == 0) {
		if ((sys = get_type ()) == NULL) {
			user_error ("You must select at least one system");
		} else
	    		Entity_status_report (sys);
	} else {
		FOR_SELECTED_ENTITY (sys)
    			Entity_status_report (sys);
		END_SELECTED_ENTITY
		free_selection ();
	}
ENDSC

SCRIPTCALLBACK(do_redraw_all_maps)
	ClassMap_redraw_all ();
ENDSC


SCRIPTCALLBACK(do_helpmode)
	set_helpMode (help_normal);
	user_inform("Now in Help Mode, use ExitHelp to get out");
ENDSC

SCRIPTCALLBACK(do_ideasmode)
	set_helpMode (help_ideas);
	user_inform("Now in Ideas Mode, use ExitHelp/Ideas to get out");
ENDSC

SCRIPTCALLBACK(do_exithelpmode)
	set_helpMode (help_none);
	user_inform("Exiting Help Mode");
ENDSC

SCRIPTCALLBACK(do_noteidea)
	add_new_idea_box ();
ENDSC

/*****************************************************************************/
/*  Init finish  */
/*****************************************************************************/

extern void MapScript_init (finish)
	MonetFinishRoutine finish;
{

	Script_init (finish);

	ScriptMacro_deeper ();
	ScriptMacro_create ( "action",	NULL,	do_action,
		"Perform a class specific action",
		"Each class of objects has an action associated with it.");
	ScriptMacro_create ( "promote-maps",	NULL,	do_free_map,
		"promote common drawings in maps",
		"Looks in the sub maps of the current map for ...");
	ScriptMacro_create ( "free-map",	NULL,	do_free_map,	"free/delete map",
		"");
	ScriptMacro_create ( "discard-map",	NULL,	do_discard_map,	"discard map",
		"Get rid of current map without saving.");
	ScriptMacro_create ( "mapset",	NULL,	do_mapset,		"set map attributes", "");
	ScriptMacro_create ( "ins",	NULL,	do_insert_drawing,	"insert <drawing> <x> <y> <width> <height>",
		"Insert/create a drawing in a map at particular position");
	ScriptMacro_create ( "list-maps",NULL,	do_list_maps,		"list loaded maps", "");
	ScriptMacro_create ( "map",	NULL,	do_change_map,		"change current map", "");
	ScriptMacro_create ( "random",	NULL,	do_random_layout,	"random layout/untangle map", "");
	ScriptMacro_create ( "layout",	NULL,	do_layout,		"layout/untangle map", "");
	ScriptMacro_create ( "localhost",NULL,	do_selectLocalHost,	"select local host", "");
	ScriptMacro_create ( "explore",	NULL,	do_explore, 		"network auto-discovery" , "");
	ScriptMacro_create ( "detach",	NULL,	do_detach, 		"detach joined entities" , "");
	ScriptMacro_create ( "join",	NULL,	do_join, 		"join drawings",
		"Link drawings together.");
	ScriptMacro_create ( "contacts",NULL,	do_contacts, 		"display entities contacts" , "");
	ScriptMacro_create ( "delete",	NULL,	do_delete,
		"delete drawings",
		"Removes all drawings in the selection list");
	ScriptMacro_create ( "delete-interfaces",	NULL,	do_delete_interfaces,
		"delete all interfaces on current map",
		"");		
	ScriptMacro_create ( "attributes",	NULL,	do_attributes, 	"list defined attribute" , "");
	ScriptMacro_create ( "shapes",	NULL,	do_shapes, 		"list available shapes" , "");
	ScriptMacro_create ( "info",	NULL,	do_info, 		"long report on entities" , "");
	ScriptMacro_create ( "list-geometry",	NULL,	do_list_geometry,
		"lists the geometry of every drawing on the current map",
		"format: <name> <label> <x> <y> <width> <height>");
	ScriptMacro_create ( "list-links",	NULL,	do_list_links,
		"lists the linked systems in the current map.",
		"any interfaces with two connections have the connections printed.");
	ScriptMacro_create ( "report",	NULL,	do_report, 		"short report on entities" , "");
	ScriptMacro_create ( "select",	NULL,	do_select, 		"select <pattern>" , "");
	ScriptMacro_create ( "stat",	NULL,	do_stat,		"status",
		"A quick summary of the state of the program");
	ScriptMacro_create ( "free",	NULL,	do_free_selection, 	"free selected" , "");
	ScriptMacro_create ( "s-list",	NULL,	do_list_selected, 	"list selected" , "");
	ScriptMacro_create ( "request-ping",NULL,	do_RequestPing,		"request ping" , "");
	ScriptMacro_create ( "clear-ping",	NULL,	do_ClearPing,		"clear ping" , "");
	ScriptMacro_create ( "list-redirect",	NULL,	do_ListRedirects,	"list query redirects",
		"Queries can be redirected to a number of agents. This command lists them.");
	ScriptMacro_create ( "reconnect",	NULL,	do_ReconnectPinger,	"reconnect monitor" , "");
	ScriptMacro_create ( "disconnect",	NULL,	do_DisconnectPinger,	"disconnect monitor" , "");
	ScriptMacro_create ( "load-edb",NULL,	do_load_db, 		"load entity database <filename>" , "");
	ScriptMacro_create ( "save-edb",NULL,	do_save_db, 		"save entity database <filename>" , "");
	ScriptMacro_create ( "put-edb",	NULL,	do_save_db_remote, 	"send entity database to monetd" , "");
	ScriptMacro_create ( "load-map",NULL,	do_load, 		"load map file <filename>" , "");
	ScriptMacro_create ( "save-map",NULL,	do_save,		"save current map" , "");
	ScriptMacro_create ( "save-maps",NULL,	do_saveall,		"save all maps" , "");
	ScriptMacro_create ( "ls-map",	NULL,	do_local_dir,		"list local maps", "");
	ScriptMacro_create ( "rls-map",	NULL,	do_remote_dir,		"list remote maps", "");
	ScriptMacro_create ( "rget-map",NULL,	do_get_remote_map,	"get remote map", "");
	ScriptMacro_create ( "rput-map",NULL,	do_put_remote_map,	"put remote maps", "");
	ScriptMacro_create ( "layout",	NULL,	do_layout, 		"layout /untangle map, <iterations>" , "");
	ScriptMacro_create ( "translate",NULL,	do_translate, 		"translate map into another class of map", "");
	ScriptMacro_create ( "lpr",	NULL,	do_lpr,			"send map to postscript printer" , "");
	ScriptMacro_create ( "dump-ps",	NULL,	do_dump_ps,		"dump map as postscript to file" , "");
	ScriptMacro_create ( "dump-text",NULL,	do_dump_text,		"dump map as text to file" , "");
	ScriptMacro_create ( "straight",NULL,	do_straighten,		"straighten interfaces" , "");
	ScriptMacro_create ( "gettop",	NULL,	do_get_selection,		"get top of selection list", "");
	ScriptMacro_create ( "currentmap",	NULL,	do_current_map,		"return name of current map" , "");
	ScriptMacro_create ( "getmonetd",NULL,	do_get_monetd,
		"return address of monetd" , "");
	ScriptMacro_create ( "getmonetmapd",NULL,do_get_monetmapd,
		"return address of monetmapd" , "");
	ScriptMacro_create ( "getedbfile",NULL,do_get_edb_file,
		"return name of default entity database file",
		"");
	ScriptMacro_create ( "set-future",NULL,	do_set_future,
		"set future status on selected list",
		"Future can be certain, temporary, planned or unknown");
	ScriptMacro_create ( "set-capacity",NULL,	do_set_capacity,
		"set capacity of entity, positive integer",
		"Depends on class of entity what this actually means");
	ScriptMacro_create ("status-report",	NULL,	do_StatusReport,
		"status report on selected list",
		"");
	ScriptMacro_create ("report-class",	NULL,	do_ReportClass,
		"status report on a class",
		"");
	ScriptMacro_create ("gather-info",	NULL,	do_gather_info,
		"gather networking information",
		"This is a bit dubious");
	ScriptMacro_create ( "redraw",	NULL,	do_redraw,		"redraw map", "");
	ScriptMacro_create ( "redraw-all",NULL,	do_redraw_all_maps,
		"redraw all maps",
		"");
	ScriptMacro_create ("select-subcomponents",	NULL,	do_SelectSubcomponents,
		"Select subcomponents of selected list",
		"");
	ScriptMacro_create ("report-subcomponents", 	NULL,	do_ReportSubcomponents,
		"Report on subcomponents of selected list",
		"");
	ScriptMacro_create ("select-supercomponents",	NULL,	do_SelectSupercomponents,
		"Select supercomponents of selected list",
		"");
	ScriptMacro_create ("report-supercomponents",	NULL,	do_ReportSupercomponents,
		"Report on supercomponents of selected list",
		"");
	ScriptMacro_create ("select-supercomponent",	NULL,	do_SelectSupercomponent,
		"Select super component",
		"This is used when making subcomponents");
	ScriptMacro_create ("make-subcomponents",	NULL,	do_MakeSubcomponent,
		"Make selected entities subcomponents",
		"Must first select a super component");
	ScriptMacro_create ("unmake-subcomponents",	NULL,	do_UnMakeSubcomponent,
		"Removes selected entities from subcomponent list",
		"Must first select a super component");
	ScriptMacro_create ("select-pingable",	NULL,	do_SelectPingable,
		"Looks in selection list for pingable entities",
		"This is dubious now");
	ScriptMacro_create ("select-pinged",	NULL,	do_SelectPinged,
		"Looks in selection list for entities being monitored",
		"This is dubious now");
	ScriptMacro_create ("helpmode",	NULL,	do_helpmode,
		"Help Mode",
		"obsolete !!!");
	ScriptMacro_create ("ideasmode",	NULL,	do_ideasmode,
		"Ideas Mode",
		"obsolete !!!");
	ScriptMacro_create ("exithelpmode",	NULL,	do_exithelpmode,
		"Exit help/ideas mode",
		"obsolete !!!");
	ScriptMacro_create ("noteidea",	NULL,	do_noteidea,
		"Note an idea",
		"obsolete !!!");
	ScriptMacro_create ("mark-map",		NULL, do_Map_mark,
		"Make current map the 'marked' map",
		"");
	ScriptMacro_create ("map-set-domain",	NULL, do_Map_SetDefaultDomain,
		"Set default domain for this map",
		"Used to shorten the labels of drawings in the map");
	ScriptMacro_create ("map-set-label",	NULL, do_Map_setLabel,
		"Set Label state for the map",
		"");
	ScriptMacro_create ("map-set-grid",	NULL, do_Map_set_grid,
		"Set drawing grid for map",
		"makes positioning lined up drawings easier");
	ScriptMacro_create ("class-map",	NULL, do_class_map,
		"Make current map a class map",
		"Map must be empty");
	ScriptMacro_create ("rename-class",	NULL, do_RenameClass,
		"Rename a class",
		"");
	ScriptMacro_create ("delete-class",	NULL, do_DeleteClass,
		"Delete a class",
		"");
	ScriptMacro_create ("select-instances",	NULL, do_SelectInstances,
		"select the instances of a class",
		"");
	ScriptMacro_create ("report-instances",	NULL, do_ReportInstances,
		"report on the instances of a class",
		"");
	ScriptMacro_create ("set-ping",	NULL,	do_SetPingOnSystem,
		"Set ping instruction for an entity instance",
		"");
	ScriptMacro_create ("set-ping-on-class",	NULL,	do_SetPingOnClass,
		"Set the ping instruction for the current class",
		"");
	ScriptMacro_create ("set-title-pattern",	NULL,	do_SetTitlePattern,
		"Set the pattern for titles for instances of a class",
		"");
	ScriptMacro_create ("rename",	NULL,	do_Rename,
		"Rename an entity",
		"");
}

extern void MapScript_finish ()
{
	ScriptMacro_shallower ();
	ScriptMacro_shallower ();
	Script_finish ();
}

/*****************************************************************************/
/*****************************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/script.c,v 1.1 1991/07/22 11:37:27 a38 Exp a38 $ */
/*****************************************************************************/
/*****************************************************************************/
