/**********************************************************/
/**********************************************************/
/* $Header$ */
/**********************************************************/
/*  Defines a set of query classes for various protocols  */
/*  Each of these protocols should be initialed here in   */
/*  'init_query_classes' which should create and return a */
/*  QueryClass structure as defined in 'query_class.h'.   */
/*  Tries to give incoming requests a high priority over existsing one so  */
/*  so that interactive users do not have to wait long.  There is a danger */
/*  of starvation if new requests come in every time around the loop.      */
/**********************************************************/

#include "query_class.h"
#include "ip.h"
#include "ip_address.h"
#include "x25.h"
#include "snmp.h"
#include "snmp_interface.h"
#include "tcp.h"

static Priority CurrentPriority = DEFAULT_PRIORITY;

/********************************************************************************/
/*  List of all query classes  */
/********************************************************************************/

static QueryClass QPList = NULL;

/**********************************************************/
/*  Convert a query into a text string */
/*    status behaviour timeofchange pingssincechange protocol address       */
/**********************************************************/

extern String Query2String (query)
	Query query;
{
	static char buffer [1024], * address;
	char * id, * status, * behaviour;
	char time[21];

	id = VIRTUAL(query) -> as_text (query);
	strncpy (time, clock2text (V_STATUS(query->entity)->time_stamp), sizeof(time));
	(void) sprintf (buffer, "%-18s %s Pings%6u %s",
		time, simple_status_report (V_STATUS(query->entity)), query->pings_since_change, id);
	if (isRedirected (query->entity))
		(void) sprintf (buffer, "%s (Redirected)", buffer);

	return buffer;
}

/********************************************************************************/
/* Make network data request (ping etc) */
/* !!!  Only does status requests, none of the other inforequests implemented */
/* !!!  yet.  */
/********************************************************************************/

static Boolean do_query (log, query)
	Query query;
	Log log;
{
	Entity entity = query->entity;
	ReachStatus reach;
	long time_stamp;
	Boolean change = False;

	time_stamp = time (NULL);
	debug_message ("time_stamp %u", time_stamp);
	debug_message ("last query %u", query->time_of_last_query);
	debug_message ("frequency %u", query->frequency);
	if (time_stamp - query->time_of_last_query < query->frequency) {
		/* -- don't query */
		return False;
	}
	VIRTUAL(query)->request(query);
	query->time_of_last_query = time_stamp;
	if (MASK(entity) & StatusReachMask) {
		query->pings_since_change ++;
	} else { /* Only record changes */
		long changesperhour;

		write_log (log, "PING", Query2String (query));
		query->change_period += time_stamp - V_STATUS(entity)->time_stamp;
		if (query->change_period > 3600*2) {
			query->change_period /= 2;
			query->changes /=2;
		}
		query->changes ++;
		query->pings_since_change = 1;
		V_STATUS(entity)->time_stamp = time_stamp;
		if (query->change_period == 0)
			changesperhour = 0;
		else
			changesperhour = query->changes / query->change_period * 3600;
		if (changesperhour > 10) {
			write_comment_log (log, "Reachability changed > 10 times in last hour");
			SetBehaviourStatusMask (V_STATUS(entity), behaviour_chaotic);
		} else if (changesperhour > 3) {
			write_comment_log (log, "Reachability changed > 3 times in last hour");
			SetBehaviourStatusMask (V_STATUS(entity), behaviour_odd);
		} else {
			SetBehaviourStatusMask (V_STATUS(entity), behaviour_normal);
		}
		change = True;
		write_log (log, "PING", Query2String (query));
	}
	debug_message ("Result: [%s]", Query2String (query));
	return change;
}


/**********************************************************/
/*  Perform queries on all managed objects  */
/*  !!! Should be more intelligent, should send back data */
/*  !!! only when it changes.  */
/*  !!! When status request is unknown, should not change */
/*  !!! the previous value, must be careful here. */
/*  !!! Need to change the protocol syntax, so that other */
/*  !!! info_requests can be returned.  */
/**********************************************************/

extern Cardinal Query_all (log)
	Log log;
{
	Query  query;
	QueryClass c;
	Cardinal TotalReachable = 0;

	debug_message ("Beginning Queries");

	TotalReachable = 0;
	/*  Do queries one class at a time  */
	FORALL (c, QPList, next) {
		if (c->enabled) {
			FORALL (query, c->instances, next_instance) {
				if (isRedirected (query->entity)) { /* !!! Should not know about redirects */
					debug_message ("Won't ping [%s] because it is redirected", Query2String (query));
				} else if (query->priority >= CurrentPriority) {
					(void) do_query (log, query);
					if (STATUS(query->entity) == status_reachable)
						TotalReachable ++;
					query -> priority = DEFAULT_PRIORITY;
				}
			}
		}
	}
	CurrentPriority = DEFAULT_PRIORITY;
	flush_log (log);
	debug_message ("Queries Done");

	debug_message ("leaving query_all_querys, total reachable [%d]", TotalReachable);
	return TotalReachable;
}

/**********************************************************/
/*  Record state of all querys  */
/**********************************************************/

extern void Query_log_all (log)
	Log log;
{
	QueryClass c;
	Query q;

	write_comment_log (log, "%s  Periodic status write out", pretty_time_stamp ());
	FORALL (c, QPList, next)
		FORALL (q, c->instances, next_instance)
			write_log (log, "PING", Query2String (q));
	write_comment_log (log, "END of write out");
	flush_log (log);
}

/********************************************************************************/
/*  Set attributes  */
/********************************************************************************/

extern void Query_set_frequency (query, frequency)
	Query query;
	TTime frequency;
{
	query -> frequency = frequency;
	debug_message ("Setting query frequency [%s] to %u",
		Entity_title (query->entity), frequency);
}

/********************************************************************************/
/*  Create and free */
/********************************************************************************/

extern Query Query_create (class_name, entity, name)
	String class_name;
	Entity entity;
	String name;
{
	Query new;
	QueryClass class;

	if ((class = String2QueryClass (class_name)) == NULL) {
		user_error ("Unknown protocol [%s] in request", class_name);
		return NULL;
	}
	if (! name || ! *name) {
		user_error ("Name of queried system is no good [%s], class [%s]",
			name, class_name);
		return NULL;
	}
	new = MonetNew (struct query);
	if (new == NULL) {
		user_error ("Could not allocate space for new managed object");
	} else {
		caddr_t subclass;
		subclass =  class->create (new, name);
		if ( ! subclass) {
			/*  Could not create subclass */
			MonetFree (new);
			return NULL;
		}
		LIST_ADD (new, class->instances, next_instance);
		new -> entity = entity;
		entity->query = new;
		new -> subclass_value = subclass;
		new -> virtual = class;
		new -> changes = 0;
		new -> change_period = 0;
		new -> pings_since_change = 0;
		new -> keep = False;
		new -> time_of_last_query = 0;
		new -> frequency = 0;
		new -> priority = VERY_HIGH_PRIORITY;
		V_STATUS(new->entity)->time_stamp = time(NULL);
		CurrentPriority = VERY_HIGH_PRIORITY;
	}
	return new;
}

extern void Query_free (query)
	Query query;
{
	VIRTUAL(query)->free(query);
	LIST_DEL (Query, query, VIRTUAL(query)->instances, next_instance);
	query->entity->query = NULL;
	MonetFree (query);
}

/****************************************************************************/
/*  Create and free */
/****************************************************************************/

extern QueryClass create_query_protocol ()
{
	QueryClass new;

	new = MonetNew (struct query_protocol);
	LIST_ADD (new, QPList, next);
	return new;
}

static void free_query_protocol (q)
	QueryClass q;
{
	q->enabled  = False;
	q->finish ();
	LIST_DEL (QueryClass, q, QPList, next);
	MonetFree (q->instances);
	MonetFree (q);
}

/****************************************************************************/
/*  Convert to ascii  */
/****************************************************************************/

extern QueryClass String2QueryClass (s)
	String s;
{
	QueryClass qp;

	FORALL (qp, QPList, next)
		if (EQS(s, qp->title))
			return qp;
	return NULL;
}

/****************************************************************************/
/*  Init and finish  */
/****************************************************************************/

extern void Query_init ()
{
	QPList = NULL;
	(void) init_snmplib ();
	(void) init_ip ();
	(void) init_ip_address ();
	(void) init_tcp ();
	(void) init_x25 ();
	(void) init_snmp ();
	(void) init_snmp_interface ();
}

extern void Query_finish ()
{
	QueryClass qp;

	FORALL (qp, QPList, next)
		free_query_protocol (qp);
	QPList = NULL;
	finish_snmplib ();
}

/**********************************************************/
/**********************************************************/
/* $Header$ */
/**********************************************************/
/**********************************************************/

